<?php

namespace Drupal\pdf_to_imagefield\Plugin\Field\FieldWidget;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\ElementInfoManagerInterface;
use Drupal\file\Plugin\Field\FieldWidget\FileWidget;
use Drupal\pdf_to_imagefield\PDFToImageManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines the 'PDF To Image Field' field widget.
 *
 * @FieldWidget(
 *   id = "pdf_to_imagefield",
 *   label = @Translation("PDF to Image"),
 *   field_types = {
 *     "file"
 *   }
 * )
 */
final class PDFToImageWidget extends FileWidget {

  /**
   * Determine valid image formats.
   *
   * @see Pdf::$validOutputFormats
   */
  const VALID_OUTPUT_FORMATS = [
    'jpg' => 'jpg',
    'jpeg' => 'jpeg',
    'png' => 'png',
  ];

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The PDFToImage manager.
   *
   * @var \Drupal\pdf_to_imagefield\PDFToImageManager
   */
  protected $pdfToImageManager;

  /**
   * {@inheritdoc}
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, array $third_party_settings, ElementInfoManagerInterface $element_info, EntityFieldManagerInterface $entityFieldManager, PDFToImageManager $pdfToImageManager) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings, $element_info);
    $this->entityFieldManager = $entityFieldManager;
    $this->pdfToImageManager = $pdfToImageManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['third_party_settings'],
      $container->get('element_info'),
      $container->get('entity_field.manager'),
      $container->get('pdf_to_imagefield.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return [
      'target_field' => NULL,
      'image_format' => 'png',
      'resolution' => 144,
      'compression' => 100,
      'hide_target_field' => FALSE,
      'only_pdf' => FALSE,
      'batch_limit' => 1,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element = parent::settingsForm($form, $form_state);

    $entity_type_id = $this->fieldDefinition->getTargetEntityTypeId();
    $bundle = $this->fieldDefinition->getTargetBundle();

    $element['description'] = [
      '#type' => 'html_tag',
      '#tag' => 'div',
      '#value' => $this->t('<strong>Important!</strong> Current and target fields must have compatibility settings: Cardinality, Allowed file extensions. Otherwise widget can cause errors.'),
    ];

    $element['target_field'] = [
      '#title' => $this->t('Image field to convert PDF into'),
      '#type' => 'select',
      '#required' => TRUE,
      '#options' => $this->pdfToImageManager->getImageTargetFields($entity_type_id, $bundle ?? $entity_type_id),
      '#default_value' => $this->getSetting('target_field'),
      '#description' => $this->t('A list of available image fields to attach PDF conversion'),
    ];

    $element['image_format'] = [
      '#title' => $this->t('Image format'),
      '#type' => 'select',
      '#required' => TRUE,
      '#options' => self::VALID_OUTPUT_FORMATS,
      '#default_value' => $this->getSetting('image_format'),
      '#description' => $this->t('Image format in which PDF pages will be created'),
    ];

    $element['resolution'] = [
      '#title' => $this->t('Resolution used for rendering PDF'),
      '#description' => $this->t('Horizontal and vertical density of the image XxY (e.g. 100x100). <em>This is not image dimensions!</em> It\'s the sampling quality of the generated image. 100x100 means the generated image will be "full size" compared to a PDF viewed at 100%. If you are only wanting preview thumbnails, you might be fine (and a LOT faster) with just 20x20 (20% size) images. Only generate full-size (100x100) images if you intend to display full-size pages. To adjust the <b>display</b> sizes on the web page, manage the display of the image field as usual. See https://www.php.net/manual/ru/imagick.setresolution.php'),
      '#type' => 'number',
      '#default_value' => $this->getSetting('resolution'),
      '#min' => 1,
    ];

    $element['compression'] = [
      '#type' => 'number',
      '#title' => $this->t('Quality of compression'),
      '#default_value' => $this->getSetting('compression'),
      '#description' => $this->t('You can set the quality of compression from 0 to 100.'),
      '#min' => 0,
      '#max' => 100,
    ];

    $element['hide_target_field'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Hide target file/image field on edit form'),
      '#default_value' => $this->getSetting('hide_target_field'),
      '#description' => $this->t('If rendering the preview image in place of the uploaded file, you may want to hide the image field from the edit form entirely to prevent editors from adding their own.'),
    ];

    $element['only_pdf'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Process only PDF files'),
      '#default_value' => $this->getSetting('only_pdf'),
      '#description' => $this->t('By default the widget converts PDF files to images and also moves other files to the target field without changes.'),
    ];

    $element['batch_limit'] = [
      '#type' => 'number',
      '#title' => $this->t('Number of PDF pages'),
      '#default_value' => $this->getSetting('batch_limit'),
      '#description' => $this->t('Number of PDF pages to process per batch operation.'),
      '#min' => 1,
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();
    $target_field = $this->getSetting('target_field');
    if (isset($target_field)) {
      $summary[] = $this->t('Attach PDF to Image field: @image_field', ['@image_field' => $target_field]);
    }
    else {
      $summary[] = $this->t('None');
    }
    $summary[] = $this->t('Image format: @image_format', ['@image_format' => $this->getSetting('image_format')]);
    $summary[] = $this->t('Resolution: @resolution', ['@resolution' => $this->getSetting('resolution')]);
    $summary[] = $this->t('Quality of compression: @compression', ['@compression' => $this->getSetting('compression')]);
    $summary[] = $this->t('Hide target field: @hide_target_field',
      ['@hide_target_field' => $this->getSetting('hide_target_field') ? $this->t('Yes') : $this->t('No')]);
    $summary[] = $this->t('Only PDF: @only_pdf',
      ['@only_pdf' => $this->getSetting('only_pdf') ? $this->t('Yes') : $this->t('No')]);
    $summary[] = $this->t('Number of PDF pages: @batch_limit', ['@batch_limit' => $this->getSetting('batch_limit')]);

    return $summary;
  }

}
